import { LabIcon } from '@jupyterlab/ui-components';

import panelSvgStr from '../style/panel.svg';

export const panelIcon = new LabIcon({
  name: '@pyviz/jupyterlab_pyviz:panel',
  svgstr: panelSvgStr
});
